/*-
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 69):
 * <syrinx@FreeBSD.org> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   -Shteryana Shopova
 * ----------------------------------------------------------------------------
 */


#include <stdlib.h>
#include "cg.h"

#define	EVENT_METHOD(i, x) GTK_WIDGET_GET_CLASS(i)->x

/* This routine gets control when the close button is clicked. */
static gboolean
cg_close_application(GtkWidget *widget, GdkEvent *event,
    gpointer data)
{
	gtk_main_quit();
	return (FALSE);
}

/* The offscreen backing pixmap and the drawing area. */
GdkPixmap *cg_pixmap = NULL;
GtkWidget *drawing_area = NULL;

/*
 * Create a new backing pixmap of the appropriate size.
 */
static gboolean
cg_configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
	if (cg_pixmap != NULL)
		g_object_unref(cg_pixmap);

	cg_pixmap = gdk_pixmap_new(widget->window,
	    widget->allocation.width, widget->allocation.height, -1);

	/* Clear the pixmap initially to white. */
	gdk_draw_rectangle(cg_pixmap, widget->style->white_gc,
	    TRUE, 0, 0, widget->allocation.width, widget->allocation.height);

	return (TRUE);
}

/* Redraw the screen from the backing pixmap. */
static gboolean
cg_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
	gdk_draw_drawable(widget->window,
	    widget->style->fg_gc[GTK_WIDGET_STATE(widget)], cg_pixmap,
	    event->area.x, event->area.y, event->area.x, event->area.y,
	    event->area.width, event->area.height);

	return (FALSE);
}

static void
cg_clear_all(GtkWidget *w, gpointer data)
{
	GtkWidget *toplevel;
	GdkRectangle update_area;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	/* Clear the pixmap initially to white. */
	gdk_draw_rectangle(cg_pixmap, toplevel->style->white_gc,
	    TRUE, 0, 0, toplevel->allocation.width,
	    toplevel->allocation.height);

	/* And update the window. */
	update_area.x = 0;
	update_area.y = 0;
	update_area.width = toplevel->allocation.width;
	update_area.height = toplevel->allocation.height;
	gtk_widget_draw(GTK_WIDGET(drawing_area), &update_area);
}

/* The colour selection dialog. */
static GtkWidget *cg_colorseldlg = NULL;

/* Color changed handler. */
void
cg_color_changed_cb(GtkWidget *widget, GtkColorSelection *colorsel)
{
	GdkColor ncolor;
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	gtk_color_selection_get_current_color (colorsel, &ncolor);
	gtk_widget_modify_fg(toplevel, GTK_STATE_NORMAL, &ncolor);
}

static void
cg_color_select(GtkWidget *w, gpointer data)
{
	GdkColor color;
	GtkWidget *toplevel;
	GtkColorSelection *colorsel;

	/* Create a color selection dialog. */
	if (cg_colorseldlg == NULL) {
		cg_colorseldlg =
		    gtk_color_selection_dialog_new ("Select colour");
	}

	/* Get the ColorSelection widget. */
	colorsel = GTK_COLOR_SELECTION(
	    GTK_COLOR_SELECTION_DIALOG(cg_colorseldlg)->colorsel);

	gtk_color_selection_set_previous_color(colorsel, &color);
	gtk_color_selection_set_current_color(colorsel, &color);
	gtk_color_selection_set_has_palette(colorsel, TRUE);

	/*
	 * Connect to the "color_changed" signal, set the client-data
	 * to the colorsel widget.
	 */
	g_signal_connect(G_OBJECT(colorsel), "color_changed",
	    G_CALLBACK(cg_color_changed_cb), (gpointer) colorsel);

	/* Show the dialog. */
	if (gtk_dialog_run(GTK_DIALOG(cg_colorseldlg)) != GTK_RESPONSE_OK) {
		toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));
		gtk_widget_modify_fg(toplevel, GTK_STATE_NORMAL, &color);	
	}

	gtk_widget_hide (cg_colorseldlg);
}

/*
 * Our menu, an array of GtkItemFactoryEntry structures that define
 * each menu item.
 */
static GtkItemFactoryEntry cg_menu_items[] = {
	{ "/Main",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Main/Select Colour",	NULL,		cg_color_select,
	   0,				"<Item>" },
	{ "/Main/_Clear All",		NULL,		cg_clear_all,
	   0,				"<StockItem>",	GTK_STOCK_DELETE },
	{ "/Main/_Quit",		NULL,		gtk_main_quit,
	   0,				"<StockItem>",	GTK_STOCK_QUIT },

	{ "/Line",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Line/Inline",		NULL,		cg_line_inline,
	   0,				"<Item>" },
	{ "/Line/Simple",		NULL,		cg_line_simple,
	   0,				"<Item>" },
	{ "/Line/Bresenhaim",		NULL,		cg_line_bresenhaim,
	   0,				"<Item>" },
	{ "/Line/Portion",		NULL,		cg_line_portion,
	   0,				"<Item>" },

	{ "/Circle",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Circle/Inline",		NULL,		cg_circle_inline,
	   0,				"<Item>" },
	{ "/Circle/Simple",		NULL,		cg_circle_simple,
	   0,				"<Item>" },
	{ "/Circle/Simple2",		NULL,		cg_circle_simple2,
	   0,				"<Item>" },
	{ "/Circle/Bresenhaim",		NULL,		cg_circle_bresenhaim,
	   0,				"<Item>" },
	{ "/Circle/Michener",		NULL,		cg_circle_michener,
	   0,				"<Item>" },
	{ "/Circle/Delta2",		NULL,		cg_circle_delta2,
	   0,				"<Item>" },

	{ "/Ellipse",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Ellipse/Input",		NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Ellipse/Input/Manual",	NULL,		cg_ellipse_set_input,
	   1,				"<RadioItem>" },
	{ "/Ellipse/Input/Auto",	NULL,		cg_ellipse_set_input,
	   2,				"/Ellipse/Input/Manual" },
	{ "/Ellipse/Inline",		NULL,		cg_ellipse_inline,
	   0,				"<Item>" },
	{ "/Ellipse/Bresenhaim",	NULL,		cg_ellipse_bresenhaim,
	   0,				"<Item>" },
	{ "/Ellipse/ConicSection",	NULL,		cg_ellipse_conic,
	   0,				"<Item>" },

	{ "/Rectangle",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Rectangle/CohenSutherland",	NULL,		cg_rectangle_cs,
	   0,				"<Item>" },
	{ "/Rectangle/LiangBarsky",	NULL,		cg_rectangle_lb,
	   0,				"<Item>" },
	{ "/Rectangle/NichollLee",	NULL,		cg_rectangle_nln,
	   0,				"<Item>" },

	{ "/Polygon",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Polygon/ScanLine",		NULL,		cg_polygon_scan_line,
	   0,				"<Item>" },
	{ "/Polygon/CyrusBeck",		NULL,		cg_polygon_cyrus_beck,
	   0,				"<Item>" },
	{ "/Polygon/SutherlandHodgman",	NULL,		cg_polygon_sh,
	   0,				"<Item>" },
	{ "/Polygon/WeilerAtherton",	NULL,		cg_polygon_wa,
	   0,				"<Item>" },

	{ "/Other",			NULL,		NULL,
	   0,				"<Branch>" },
	{ "/Other/Spline",		NULL,		cg_other_spline,
	   0,				"<Item>" },
	{ "/Other/Holiday",		NULL,		cg_other_holiday,
	   0,				"<Item>" },
	{ "/Other/Overhauser",		NULL,		cg_other_overhauser,
	   0,				"<Item>" },
	{ "/Other/B-Spline",		NULL,		cg_other_b_spline,
	   0,				"<Item>" },
	{ "/Other/Besier",		NULL,		cg_other_besier,
	   0,				"<Item>" },
	{ "/Other/Ferguson",		NULL,		cg_other_ferguson,
	   0,				"<Item>" },
	{ "/_Help",			NULL,		NULL,
	   0,				"<LastBranch>" },
	{ "/_Help/About",		NULL,		NULL,
	   0,				"<Item>" },
};

static gint cg_nitems = sizeof(cg_menu_items) / sizeof(cg_menu_items[0]);

/* Create the menubar. */
static GtkWidget *
cg_create_menubar(void)
{
	GtkItemFactory *item_factory;

	/* Make an ItemFactory (that makes a menubar). */
	item_factory = gtk_item_factory_new(GTK_TYPE_MENU_BAR, "<main>", NULL);

	/*
	 * This function generates the menu items. Pass the item factory,
	 * the number of items in the array, the array itself, and any
	 * callback data for the the menu items.
	 */
	gtk_item_factory_create_items(item_factory, cg_nitems,
	    cg_menu_items, NULL);

	/* Finally, return the actual menu bar created by the item factory. */
	return (gtk_item_factory_get_widget(item_factory, "<main>"));
}

#define	XSIZE 		800 /* 1400 */
#define	YSIZE		600 /* 900 */

/* Create a ruler and pack it in a table. */
static void
cg_create_ruler(GtkWidget *table, GtkWidget *area)
{
	GtkWidget *hrule, *vrule;

	/*
	 * The horizontal ruler goes on top. As the mouse moves across the
	 * drawing area, a motion_notify_event is passed to the
	 * appropriate event handler for the ruler.
	 */
	hrule = gtk_hruler_new();
	gtk_ruler_set_metric(GTK_RULER(hrule), GTK_PIXELS);
	gtk_ruler_set_range(GTK_RULER(hrule), -XSIZE, XSIZE, 10, 2 * XSIZE);

	g_signal_connect_swapped(G_OBJECT(area), "motion_notify_event",
	    G_CALLBACK(EVENT_METHOD(hrule, motion_notify_event)),
	    G_OBJECT(hrule));

	gtk_table_attach(GTK_TABLE(table), hrule, 1, 2, 0, 1,
	    GTK_EXPAND | GTK_SHRINK | GTK_FILL, GTK_FILL, 0, 0);

	/*
	 * The vertical ruler goes on the left. As the mouse moves across
	 * the drawing area, a motion_notify_event is passed to the
	 * appropriate event handler for the ruler.
	 */
	vrule = gtk_vruler_new();
	gtk_ruler_set_metric(GTK_RULER(vrule), GTK_PIXELS);
	gtk_ruler_set_range(GTK_RULER(vrule), YSIZE, -YSIZE, 10, 2 * YSIZE);

	g_signal_connect_swapped(G_OBJECT(area), "motion_notify_event",
	    G_CALLBACK(EVENT_METHOD(vrule, motion_notify_event)),
	    G_OBJECT (vrule));

	gtk_table_attach(GTK_TABLE(table), vrule, 0, 1, 1, 2,
	    GTK_FILL, GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0);

	/* We are now ready to display the rulers. */
	gtk_widget_show (hrule);
	gtk_widget_show (vrule);
}

GtkWidget *window;
/* The main routine. */
int
main(int argc, char *argv[])
{
	GtkWidget *window, *main_vbox, *menubar, *table;
	GtkWidget *separator;

	/* Initialize GTK and create the main window. */
	gtk_init(&argc, &argv);

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	g_signal_connect(G_OBJECT(window), "delete_event",
	    G_CALLBACK(cg_close_application), NULL);
	gtk_container_set_border_width(GTK_CONTAINER(window), 5);

	/* Make a vbox to put the menubar and table. */
	main_vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_set_border_width(GTK_CONTAINER(main_vbox), 1);
	gtk_container_add(GTK_CONTAINER(window), main_vbox);

	/* Add the menubar. */
	menubar = cg_create_menubar();
	gtk_box_pack_start(GTK_BOX(main_vbox), menubar, FALSE, TRUE, 0);
	gtk_widget_show(menubar);

	/* Add a separator. */
	separator = gtk_hseparator_new();
	gtk_box_pack_start(GTK_BOX(main_vbox), separator, FALSE, TRUE, 5);
	gtk_widget_show(separator);

	/* Create a table for placing the ruler and the drawing area. */
	table = gtk_table_new(3, 2, FALSE);
	gtk_box_pack_end(GTK_BOX(main_vbox), table, TRUE, TRUE, 0);

	drawing_area = gtk_drawing_area_new();
	gtk_widget_set_size_request(GTK_WIDGET(drawing_area), XSIZE, YSIZE);
	gtk_table_attach(GTK_TABLE(table), drawing_area, 1, 2, 1, 2,
	    GTK_EXPAND | GTK_FILL , GTK_FILL, 0, 0);

	/* Signals used to handle backing pixmap. */
	g_signal_connect(G_OBJECT(drawing_area), "expose_event",
	    G_CALLBACK(cg_expose_event), NULL);

	g_signal_connect(G_OBJECT(drawing_area), "configure_event",
	    G_CALLBACK(cg_configure_event), NULL);

	gtk_widget_set_events (drawing_area, GDK_EXPOSURE_MASK
	    | GDK_LEAVE_NOTIFY_MASK | GDK_BUTTON_PRESS_MASK
	    | GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK
	    | GDK_POINTER_MOTION_HINT_MASK);

	gtk_widget_show (drawing_area);

	/* Create the rulers and display them. */
	cg_create_ruler(table, drawing_area);

	/* Now show everything else.*/
	gtk_widget_show(table);
	gtk_widget_show(main_vbox);
	gtk_widget_show(window);

	/* Rest in gtk_main and wait for the fun to begin! */
	gtk_main();

	return (0);
}
