/*-
 * ----------------------------------------------------------------------------
 * "THE BEER-WARE LICENSE" (Revision 69):
 * <syrinx@FreeBSD.org> wrote this file.  As long as you retain this notice you
 * can do whatever you want with this stuff. If we meet some day, and you think
 * this stuff is worth it, you can buy me a beer in return.   -Shteryana Shopova
 * ----------------------------------------------------------------------------
 */
#include <sys/queue.h>

#include <math.h>
#include <stdlib.h>
#include <strings.h>

#include "cg.h"

#define	MAX_EDGES	50

/* A polygon. */
struct cg_polygon {
	gint	x_[MAX_EDGES + 1];
	gint	y_[MAX_EDGES + 1];
	gint	n_edge;
	gint	c_edges;
} polygons[2];

static GtkWidget *dialog;
static GtkWidget *t_entry;
gint how_many;
gint poly_n;
gulong	poly_button_id;	/*  button press/release handler id */
gulong	poly_motion_id;	/*  mouse motion handler id */

typedef void (*cg_polygon_next_cb) (void);

cg_polygon_next_cb next_cb;

static void cg_polygon_input_nedges(void);

static void
cg_polygon_clear_line(void)
{
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    polygons[poly_n].x_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].y_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].x_[polygons[poly_n].c_edges],
	    polygons[poly_n].y_[polygons[poly_n].c_edges]);

	cg_line_drawing_refresh(polygons[poly_n].x_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].y_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].x_[polygons[poly_n].c_edges],
	    polygons[poly_n].y_[polygons[poly_n].c_edges]);
}

static void
cg_polygon_draw_line(GtkStateType w_state)
{
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[w_state],
	    polygons[poly_n].x_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].y_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].x_[polygons[poly_n].c_edges],
	    polygons[poly_n].y_[polygons[poly_n].c_edges]);

	cg_line_drawing_refresh(polygons[poly_n].x_[polygons[poly_n].c_edges  - 1],
	    polygons[poly_n].y_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].x_[polygons[poly_n].c_edges],
	    polygons[poly_n].y_[polygons[poly_n].c_edges]);
}

static void
cg_polygon_draw_last_line(GtkStateType w_state)
{
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[w_state],
	    polygons[poly_n].x_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].y_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].x_[0],
	    polygons[poly_n].y_[0]);

	cg_line_drawing_refresh(polygons[poly_n].x_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].y_[polygons[poly_n].c_edges - 1],
	    polygons[poly_n].x_[0],
	    polygons[poly_n].y_[0]);

	polygons[poly_n].x_[polygons[poly_n].c_edges] = polygons[poly_n].x_[0];
	polygons[poly_n].y_[polygons[poly_n].c_edges] = polygons[poly_n].y_[0];
}

gint
cg_polygon_motion_notify_event(GtkWidget *widget, GdkEventButton *event)
{
	cg_polygon_clear_line();
	polygons[poly_n].x_[polygons[poly_n].c_edges] = event->x;
	polygons[poly_n].y_[polygons[poly_n].c_edges] = event->y;
	cg_polygon_draw_line(GTK_WIDGET_STATE(widget));

	return (TRUE);
}

gint
cg_polygon_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
	if (cg_pixmap == NULL || event->button != 1)
		return (TRUE);

	if (polygons[poly_n].c_edges + 1 == polygons[poly_n].n_edge) {
		g_signal_handler_disconnect(GTK_OBJECT(drawing_area),
		    poly_motion_id);
		g_signal_handler_disconnect(GTK_OBJECT(drawing_area),
		    poly_button_id);

		cg_polygon_clear_line();
		cg_polygon_draw_line(GTK_WIDGET_STATE(widget));
		polygons[poly_n].c_edges++;
		cg_polygon_draw_last_line(GTK_WIDGET_STATE(widget));

		if (how_many > 1 && poly_n == 0) {
			poly_n++;
			cg_polygon_input_nedges();
		} else
			next_cb();

		return (TRUE);
	}

	polygons[poly_n].x_[polygons[poly_n].c_edges] = event->x;
	polygons[poly_n].y_[polygons[poly_n].c_edges] = event->y;

	if (polygons[poly_n].c_edges == 0) {
		poly_motion_id = g_signal_connect(GTK_OBJECT(drawing_area),
		    "motion_notify_event",
		    (GtkSignalFunc) cg_polygon_motion_notify_event, NULL);
	} else {
		cg_polygon_clear_line();
		cg_polygon_draw_line(GTK_WIDGET_STATE(widget));
	}

	polygons[poly_n].c_edges++;

	return (TRUE);
}


static void
cg_polygon_draw(gint which, gint n_edge)
{
	if (which < 0 || which > 1)
		return;

	bzero(polygons + which, sizeof(struct cg_polygon));
	polygons[which].n_edge = n_edge;
	polygons[which].c_edges = 0;

	poly_button_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "button_press_event",
	    (GtkSignalFunc) cg_polygon_button_press_event, NULL);
}

static void
cg_polygon_check_nedges(GtkWidget *widget, GtkWidget *entry)
{
	gint n_edge;

	if ((n_edge = strtoul(gtk_entry_get_text(GTK_ENTRY(t_entry)),
	    NULL, 10 )) > MAX_EDGES || n_edge < 4)
		return;

	gtk_widget_hide(dialog);

	cg_polygon_draw(poly_n, n_edge);
}

static void
cg_polygon_input_nedges(void)
{
	GtkWidget *button, *label;

	dialog = gtk_dialog_new ();
	g_signal_connect(G_OBJECT(dialog), "destroy",
	    G_CALLBACK(gtk_widget_hide), NULL);

	gtk_container_set_border_width(GTK_CONTAINER(dialog), 0);

	label = gtk_label_new("How many edges? [4 - 50] \n");
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), label, FALSE, FALSE, 0);
	gtk_widget_show(label);

	t_entry = gtk_entry_new_with_max_length(2);
	gtk_entry_set_text(GTK_ENTRY(t_entry), "5");
	gtk_editable_set_editable(GTK_EDITABLE(t_entry), TRUE);
	gtk_entry_set_visibility(GTK_ENTRY(t_entry), TRUE);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), t_entry, FALSE, FALSE, 0);
	gtk_widget_show(t_entry);

	button = gtk_button_new_from_stock(GTK_STOCK_OK);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), button, TRUE, TRUE, 0);
	g_signal_connect_swapped (G_OBJECT (button), "clicked",
	    G_CALLBACK (cg_polygon_check_nedges), G_OBJECT(dialog));
	gtk_widget_show(button);

	/* This makes the "OK" button default. */
	GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT);
	gtk_widget_grab_default(button);
	gtk_widget_show(button);

	gtk_widget_show (dialog);
}

struct pg_seg {
	gint		index;
	gint		y_min;
	gint		y_max;
	gdouble		x_min;
	gdouble		inc_x;
	gboolean	x_decr;
	TAILQ_ENTRY(pg_seg) link;
};

TAILQ_HEAD(sorted_seg_list, pg_seg);

struct pg_seg_y_axis {
	gint	y_;
	gint	num_e;
	struct sorted_seg_list	   tr_segs;
	TAILQ_ENTRY(pg_seg_y_axis) y_axis;
};

TAILQ_HEAD(tr_struct, pg_seg_y_axis);

struct tr_struct tr_list;
struct sorted_seg_list active_segments;
struct sorted_seg_list all_segments;

struct cg_polygon *sl_poly = &polygons[0];

static gint
cg_poly_fill_sl_find_x_min(void)
{
	gint i, x_min;

	x_min = sl_poly->x_[0];

	for (i = 1; i < sl_poly->n_edge; i++) {
		if (sl_poly->x_[i] < x_min)
			x_min = sl_poly->x_[i];
	}

	return (x_min);
}

static gint
cg_poly_fill_sl_find_x_max(void)
{
	gint i, x_max;

	x_max = sl_poly->x_[0];

	for (i = 1; i < sl_poly->n_edge; i++) {
		if (sl_poly->x_[i] > x_max)
			x_max = sl_poly->x_[i];
	}

	return (x_max);
}

static gint
cg_poly_fill_sl_find_y_min(void)
{
	gint i, y_min;

	y_min = sl_poly->y_[0];

	for (i = 1; i < sl_poly->n_edge; i++) {
		if (sl_poly->y_[i] < y_min)
			y_min = sl_poly->y_[i];
	}

	return (y_min);
}

static gint
cg_poly_fill_sl_find_y_max(void)
{
	gint i, y_max;

	y_max = sl_poly->y_[0];

	for (i = 1; i < sl_poly->n_edge; i++) {
		if (sl_poly->y_[i] > y_max)
			y_max = sl_poly->y_[i];
	}

	return (y_max);
}

static struct pg_seg *
cg_sl_segment_prepare(gint n_p)
{
	gdouble x_max;
	struct pg_seg *new_seg;

	if (sl_poly->y_[n_p] == sl_poly->y_[n_p + 1])
		return (NULL);

	if ((new_seg = malloc(sizeof(struct pg_seg))) == NULL)
		abort();

	bzero(new_seg, sizeof(struct pg_seg));
	new_seg->index = n_p;

	if (sl_poly->x_[n_p + 1] - sl_poly->x_[n_p] < 0)
		new_seg->x_decr = TRUE;

	if (sl_poly->y_[n_p] > sl_poly->y_[n_p + 1]) {
		new_seg->y_min = sl_poly->y_[n_p + 1];
		new_seg->x_min = sl_poly->x_[n_p + 1];
		new_seg->y_max = sl_poly->y_[n_p];
		x_max = sl_poly->x_[n_p];
	} else {
		new_seg->y_min = sl_poly->y_[n_p];
		new_seg->x_min = sl_poly->x_[n_p];
		new_seg->y_max = sl_poly->y_[n_p + 1];
		x_max = sl_poly->x_[n_p + 1];
	}

	new_seg->inc_x = (x_max - new_seg->x_min) /
	    (new_seg->y_max - new_seg->y_min);

	return (new_seg);
}

static void
cg_sl_build_segments(void)
{
	gint i;
	struct pg_seg *next_seg;

	bzero(&all_segments, sizeof(struct sorted_seg_list));

	all_segments.tqh_first = NULL;
	all_segments.tqh_last = &(all_segments).tqh_first;

	for (i = 0; i < sl_poly->c_edges; i++) {
		if ((next_seg = cg_sl_segment_prepare(i)) == NULL)
			continue;

		TAILQ_INSERT_TAIL(&all_segments, next_seg, link);
	}
}

static void
cg_sl_insert_sorted_segs(struct sorted_seg_list *head_p,
    struct pg_seg *new_en)
{
	struct pg_seg *next;

	TAILQ_FOREACH(next, head_p, link) {
		if (new_en->x_min < next->x_min) {
			TAILQ_INSERT_BEFORE(next, new_en, link);
			return;
		}
	}

	TAILQ_INSERT_TAIL(head_p, new_en, link);
}

static void
cg_sl_build_sorted_segs(struct pg_seg_y_axis *eq_y)
{
	struct pg_seg *cur, *temp;

	TAILQ_FOREACH_SAFE(cur, &all_segments, link, temp) {
		if (cur->y_min == eq_y->y_) {
			TAILQ_REMOVE(&all_segments, cur, link);
			cg_sl_insert_sorted_segs(&(eq_y->tr_segs), cur);
			eq_y->num_e++;
		}
	}
}

static void
cg_sl_build_table(void)
{
	gint i, n_y, y_min;
	struct pg_seg_y_axis *p_y;

	bzero(&tr_list, sizeof(struct tr_struct));
	tr_list.tqh_first = NULL;
	tr_list.tqh_last = &(tr_list).tqh_first;

	cg_sl_build_segments();
	y_min = cg_poly_fill_sl_find_y_min();
	n_y = cg_poly_fill_sl_find_y_max() - y_min;

	for (i = 0; i < n_y; i++, y_min++) {
		if ((p_y = malloc(sizeof(struct pg_seg_y_axis))) == NULL)
			abort();

		bzero(p_y, sizeof(struct pg_seg_y_axis));
		p_y->tr_segs.tqh_first = NULL;
		p_y->tr_segs.tqh_last = &(p_y->tr_segs).tqh_first;
		p_y->y_ = y_min;
		p_y->num_e = 0;

		cg_sl_build_sorted_segs(p_y);
		TAILQ_INSERT_TAIL(&tr_list, p_y, y_axis);
	}
}

static void
cg_poly_draw_line(struct pg_seg *curr_seg, struct pg_seg * temp_seg,
    gint y_)
{
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	gdk_draw_line(cg_pixmap, toplevel->style->mid_gc[GTK_STATE_NORMAL],
	    curr_seg->x_min, y_, temp_seg->x_min, y_);
}

void
cg_polygon_fill_scan_line(void)
{
	gint x_min, y_min, x_max, y_max, last_y;

	struct pg_seg *curr_seg, *temp_seg;
	struct pg_seg_y_axis *curr_y;
	struct sorted_seg_list temp_active_segments;

	sl_poly = polygons;

	x_min = cg_poly_fill_sl_find_x_min();
	x_max = cg_poly_fill_sl_find_x_max();
	y_min = cg_poly_fill_sl_find_y_min();
	y_max = cg_poly_fill_sl_find_y_max();

	/* 1. Postroqvame tablica na rebrata */
	cg_sl_build_table();

	/* 2. Initializirane spisaka ot acttivni rebra na prazen. */
	bzero(&active_segments, sizeof(active_segments));
	active_segments.tqh_first = NULL;
	active_segments.tqh_last = &(active_segments).tqh_first;

	last_y = y_min;
	/* 3. zapochvame ot y_min */
	for ( ; last_y < y_max; last_y++)  {
		if ((curr_y = TAILQ_FIRST(&tr_list)) != NULL) {
			/* 4a. Dobavqme kym SAR spisyka ot TR za y_. */
			TAILQ_CONCAT(&active_segments, &(curr_y->tr_segs), link);
		}

		/*
		 * 4b zapisvame pixelite mezhdu vsqka dvojka 
		 * nechetna - chetna prese4ni to4ki or SAR
		 */

		TAILQ_FOREACH(curr_seg, &active_segments, link) {
			if ((temp_seg = TAILQ_NEXT(curr_seg, link)) == NULL)
				break;

			cg_poly_draw_line(curr_seg, temp_seg, last_y);
			curr_seg = temp_seg;
		}

		/* 4c. y_++ Iztrivame os SAR vsi4ki rebra za koito y_max == y_ */
		TAILQ_FOREACH_SAFE(curr_seg, &active_segments, link, temp_seg) {
			//if (curr_seg->x_decr)
			//	curr_seg->x_min -= curr_seg->inc_x;
			//else
				curr_seg->x_min += curr_seg->inc_x;

			if (curr_seg->y_max <= last_y) {
				TAILQ_REMOVE(&active_segments, curr_seg, link);
				free(curr_seg);
			}
		}

		bzero(&temp_active_segments, sizeof(temp_active_segments));
		temp_active_segments.tqh_first = NULL;
		temp_active_segments.tqh_last = &(temp_active_segments).tqh_first;

		TAILQ_CONCAT(&temp_active_segments, &active_segments, link);

		while ((curr_seg = TAILQ_FIRST(&temp_active_segments)) != NULL) {
			TAILQ_REMOVE(&temp_active_segments, curr_seg, link);
			cg_sl_insert_sorted_segs(&active_segments, curr_seg);
		}

		if (curr_y != NULL) {
			TAILQ_REMOVE(&tr_list, curr_y, y_axis);
			free(curr_y);
		}

		if (TAILQ_EMPTY(&active_segments))
			break;
	}

	cg_line_drawing_refresh(x_min, y_min, x_max, y_max);
}

static struct cg_segment_line cb_seg = { 0, 0, 0, 0 };

#define	DT_PROD(x1, y1, x2, y2)	\
	((x1) * (x2) + (y1)  * (y2))

static void
cg_polygon_clip_cyrus_beck(GtkWidget *widget)
{
	gint i;
	gdouble	tin = 0, tout = 1, xn, yn, xv, yv, nv, nd;
	GtkWidget *toplevel;

	for (i = 0; i < polygons[0].n_edge; i++) {
		xn = polygons[0].y_[i + 1] - polygons[0].y_[i];
		yn = polygons[0].x_[i] - polygons[0].x_[i + 1];
		xv = cb_seg.x1_ - polygons[0].x_[i];
		yv = cb_seg.y1_ - polygons[0].y_[i];

		nv = DT_PROD(xn, yn, xv, yv);
		nd = DT_PROD(xn, yn, (cb_seg.x2_ - cb_seg.x1_), (cb_seg.y2_ - cb_seg.y1_));

		if (nd > 0) {
			if  (tin < (-nv/nd))
				tin = (-nv/nd);
		} else if (nd < 0) {
			if (tout > (-nv/nd))
				tout = (-nv/nd);
		} else if (nv < 0)
			return;
	}

	if (tin > tout)
		return;

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	gdk_draw_line(cg_pixmap, toplevel->style->mid_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_ + tin * (cb_seg.x2_ - cb_seg.x1_), cb_seg.y1_ + tin * (cb_seg.y2_ - cb_seg.y1_),
	    cb_seg.x1_ + tout * (cb_seg.x2_ - cb_seg.x1_), cb_seg.y1_ + tout * (cb_seg.y2_ - cb_seg.y1_));

	cg_line_drawing_refresh(cb_seg.x1_ + tin * (cb_seg.x2_ - cb_seg.x1_), cb_seg.y1_ + tin * (cb_seg.y2_ - cb_seg.y1_),
	    cb_seg.x1_ + tout * (cb_seg.x2_ - cb_seg.x1_), cb_seg.y1_ + tout * (cb_seg.y2_ - cb_seg.y1_));
}

#define IS_VISIBLE(px, py, x1, y1, x2, y2)	\
	(((py) - (y1)) * ((x2) - (x1)) <= ((px) - (x1)) * ((y2)  - (y1))) 

static void
cg_intersect_x(gint *x_res, gint *y_res, gint x_cnst, gint x1_, gint x2_, gint y1_, gint y2_)
{
	if (x1_ == x2_) {
		g_message ("x1=x2 => x || x_cnst - cannot intersect\n");
		return;
	}

	*x_res = x_cnst;
	*y_res = y1_ + ((y2_ - y1_) * (x_cnst  - x1_))/(x2_- x1_);
}

static void
cg_intersect_y(gint *x_res, gint *y_res, gint y_cnst, gint x1_, gint x2_, gint y1_, gint y2_)
{
	if (y2_ == y1_) {
		g_message ("y1=y2 => y || y_cnst - cannot intersect\n");
		return;
	}

	*x_res = x1_ + ((x2_ - x1_) * (y_cnst  - y1_))/(y2_- y1_);
	*y_res = y_cnst;
}

typedef gboolean (*cg_visible_cb) (gint x1_, gint y1_, gint x_min, gint y_min, gint x_max, gint y_max);

static gint
cg_poly_clip_sh_line(gint x1_, gint y1_, gint x2_, gint y2_,
    struct cg_polygon *poly_in, struct cg_polygon *poly_out,
    cg_visible_cb isVis, gboolean *isVisLast)
{
	gboolean is_visible1, is_visible2, x_cnst;
	gint i, x, y;

	bzero(poly_out, sizeof(struct cg_polygon));

	if (x2_ == x1_)
		x_cnst = 1;
	else
		x_cnst = 0;

	is_visible1 = isVis(poly_in->x_[poly_in->n_edge - 1],
	    poly_in->y_[poly_in->n_edge - 1], x1_, y1_, x2_, y2_);

	x = poly_in->x_[poly_in->n_edge - 1];
	y = poly_in->y_[poly_in->n_edge - 1];
	poly_out->n_edge = 0;

#if 0
	if (!(*isVisLast) && !is_visible1) {
		/*
		 * We need to add the vertex of the rectangle to
		 * the resulting polygon.
		 */
		poly_out->x_[0] = x1_;
		poly_out->y_[0] = y1_;
		poly_out->n_edge++;
	}
#endif
	for (i = 0; i < poly_in->n_edge; i++) {
		is_visible2 = isVis(poly_in->x_[i], poly_in->y_[i], x1_, y1_, x2_, y2_);

		if (is_visible1 && is_visible2) {
			/* Both ends visible - add only second. */
			poly_out->x_[poly_out->n_edge] = poly_in->x_[i];
			poly_out->y_[poly_out->n_edge] = poly_in->y_[i];
			poly_out->n_edge++;
		} else if (is_visible1 && !is_visible2) {
			/*
			 * The line left the visible part of the plane -
			 * add the intersection point only.
			 */
			if (x_cnst)
				cg_intersect_x(&(poly_out->x_[poly_out->n_edge]),
				    &(poly_out->y_[poly_out->n_edge]), x1_, x,
				    poly_in->x_[i], y, poly_in->y_[i]);
			else
				cg_intersect_y(&(poly_out->x_[poly_out->n_edge]),
				    &(poly_out->y_[poly_out->n_edge]), y1_, x,
				    poly_in->x_[i], y, poly_in->y_[i]);

			poly_out->n_edge++;
		} else if (!is_visible1 && is_visible2) {
			/*
			 * The line enters the visible part of the plane - 
			 * add the intersection point and second vertex.
			 */
			if (x_cnst)
				cg_intersect_x(&(poly_out->x_[poly_out->n_edge]),
				    &(poly_out->y_[poly_out->n_edge]), x1_, x,
				    poly_in->x_[i], y, poly_in->y_[i]);
			else
				cg_intersect_y(&(poly_out->x_[poly_out->n_edge]),
				    &(poly_out->y_[poly_out->n_edge]), y1_, x,
				    poly_in->x_[i], y, poly_in->y_[i]);
			poly_out->n_edge++;

			poly_out->x_[poly_out->n_edge] = poly_in->x_[i];
			poly_out->y_[poly_out->n_edge] = poly_in->y_[i];
			poly_out->n_edge++;
		}
		is_visible1 = is_visible2;
		x = poly_in->x_[i];
		y = poly_in->y_[i];
	}

	poly_out->c_edges = poly_out->n_edge;

	*isVisLast = is_visible2;

	return (poly_out->n_edge);
}

static gboolean
cg_is_visible(gint x_, gint y_, gint x1_, gint y1_, gint x2_, gint y2_)
{
	return ((y_ - y1_) * (x2_ - x1_) > (x_ - x1_) * (y2_ - y1_));
}


static void
cg_polygon_clip_sh(GtkWidget *widget)
{
	gboolean isVisLast = 1;
	gint i;
	gdouble x_min, x_max, y_min, y_max;
	struct cg_polygon polys[4];

	if (cb_seg.x1_ > cb_seg.x2_) {
		x_min = cb_seg.x2_;
		x_max = cb_seg.x1_;
	} else if (cb_seg.x1_ < cb_seg.x2_) {
		x_min = cb_seg.x1_;
		x_max = cb_seg.x2_;
	} else
		return;

	if (cb_seg.y1_ < cb_seg.y2_) {
		y_min = cb_seg.y1_;
		y_max = cb_seg.y2_;
	} else if (cb_seg.y1_ > cb_seg.y2_) {
		y_min = cb_seg.y2_;
		y_max = cb_seg.y1_;
	} else
		return;

	bzero(polys, 4 * sizeof(struct cg_polygon));
	bzero(polygons, sizeof(struct cg_polygon));

	if (cg_poly_clip_sh_line(x_min, y_min, x_max, y_min, polygons + 1, polys, cg_is_visible, &isVisLast) == 0)
		return;

	if (cg_poly_clip_sh_line(x_max, y_min, x_max, y_max, polys, polys + 1, cg_is_visible, &isVisLast) == 0)
		return;

	if (cg_poly_clip_sh_line(x_max, y_max, x_min, y_max, polys + 1, polys + 2, cg_is_visible, &isVisLast) == 0)
		return;

	if (cg_poly_clip_sh_line(x_min, y_max, x_min, y_min, polys + 2, polygons, cg_is_visible, &isVisLast) == 0) 
		return;

	/* sl_poly = polys + 3; */

	for (i = 0; i < polys[3].n_edge; i++) {
		g_message ("(%d, %d)!\n", polys[3].x_[i], polys[3].y_[i]);
	}

	cg_polygon_fill_scan_line();
}

static void
cg_polygon_clip_wa(void)
{
}

void
cg_polygon_scan_line(GtkWidget *w, gpointer data)
{
	g_message ("Hello, cg_polygon_scan_line!\n");
	poly_n = 0;
	how_many = 1;
	next_cb = cg_polygon_fill_scan_line;
	cg_polygon_input_nedges();
}

static gint
cg_poly_cb_motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
	int x, y;
	GdkModifierType state;
	GtkWidget *toplevel;

	if (event->is_hint)
		gdk_window_get_pointer (event->window, &x, &y, &state);
	else {
		x = event->x;
		y = event->y;
		state = event->state;
	}

	if (cg_pixmap == NULL || (state & GDK_BUTTON1_MASK) == 0)
		return (TRUE);

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	/* Clear previous line. */
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cb_seg.x2_ = x;
	cb_seg.y2_ = y;

	/* Draw the new line. */
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);
	/* And clip it. */
	/* cg_polygon_clip_cyrus_beck(widget); */

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	return (TRUE);
}

static gint
cg_poly_cb_button_release_event(GtkWidget *widget, GdkEventButton *event)
{
	GtkWidget *toplevel;

	if (cg_pixmap == NULL || event->button != 1)
		return (TRUE);

	/* Unconnect the motion and the button release signals. */
	g_signal_handler_disconnect(GTK_OBJECT(drawing_area), poly_motion_id);
	g_signal_handler_disconnect(GTK_OBJECT(drawing_area), poly_button_id);

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	/* Clear previous line. */
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cb_seg.x2_ = event->x;
	cb_seg.y2_ = event->y;

	/* Draw the new line. */
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);
	/* And clip it. */
	cg_polygon_clip_cyrus_beck(widget);

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	return (TRUE);
}

static gint
cg_poly_cb_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
	if (cg_pixmap == NULL || event->button != 1)
		return (TRUE);

	cb_seg.x1_ = event->x;
	cb_seg.y1_ = event->y;

	/* Unconnect the button press signal. */
	g_signal_handler_disconnect(GTK_OBJECT(drawing_area), poly_button_id);

	/* Connect the motion and the button release signal handler. */
	poly_button_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "button_release_event",
	    (GtkSignalFunc) cg_poly_cb_button_release_event, NULL);

	poly_motion_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "motion_notify_event",
	    (GtkSignalFunc) cg_poly_cb_motion_notify_event, NULL);


	return (TRUE);
}

static void
cg_poly_cb_input_segment(void)
{
	/* Input the line segment. */
	poly_button_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "button_press_event",
	    (GtkSignalFunc) cg_poly_cb_button_press_event, NULL);
}

void
cg_polygon_cyrus_beck(GtkWidget *w, gpointer data)
{
	g_message ("Hello, cg_polygon_cyrus_beck!\n");
	poly_n = 0;
	how_many = 1;
	next_cb = cg_poly_cb_input_segment;
	cg_polygon_input_nedges();
}

static gint
cg_poly_sh_motion_notify_event(GtkWidget *widget, GdkEventMotion *event)
{
	int x, y;
	GdkModifierType state;
	GtkWidget *toplevel;

	if (event->is_hint)
		gdk_window_get_pointer (event->window, &x, &y, &state);
	else {
		x = event->x;
		y = event->y;
		state = event->state;
	}

	if (cg_pixmap == NULL || (state & GDK_BUTTON1_MASK) == 0)
		return (TRUE);

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	/* Clear previous rectangle. */
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y1_);
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x1_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y2_, cb_seg.x2_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x2_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cb_seg.x2_ = event->x;
	cb_seg.y2_ = event->y;

	/* Draw the new rectangle. */
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y1_);
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x1_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y2_, cb_seg.x2_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x2_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);
	/* And clip it. */
	/* cg_polygon_clip_cyrus_beck(widget); */

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	return (TRUE);
}

static gint
cg_poly_sh_button_release_event(GtkWidget *widget, GdkEventButton *event)
{
	GtkWidget *toplevel;

	if (cg_pixmap == NULL || event->button != 1)
		return (TRUE);

	/* Unconnect the motion and the button release signals. */
	g_signal_handler_disconnect(GTK_OBJECT(drawing_area), poly_motion_id);
	g_signal_handler_disconnect(GTK_OBJECT(drawing_area), poly_button_id);

	toplevel = gtk_widget_get_toplevel(GTK_WIDGET(drawing_area));

	/* Clear previous rectangle. */
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y1_);
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x1_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x1_, cb_seg.y2_, cb_seg.x2_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->white_gc,
	    cb_seg.x2_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	cb_seg.x2_ = event->x;
	cb_seg.y2_ = event->y;

	/* Draw the new rectangle. */
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y1_);
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y1_, cb_seg.x1_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x1_, cb_seg.y2_, cb_seg.x2_, cb_seg.y2_);
	gdk_draw_line(cg_pixmap, toplevel->style->fg_gc[GTK_WIDGET_STATE(widget)],
	    cb_seg.x2_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);
	/* And clip it. */
	cg_polygon_clip_sh(widget);

	cg_line_drawing_refresh(cb_seg.x1_, cb_seg.y1_, cb_seg.x2_, cb_seg.y2_);

	return (TRUE);
}

static gint
cg_poly_sh_button_press_event(GtkWidget *widget, GdkEventButton *event)
{
	if (cg_pixmap == NULL || event->button != 1)
		return (TRUE);

	cb_seg.x1_ = event->x;
	cb_seg.y1_ = event->y;

	/* Unconnect the button press signal. */
	g_signal_handler_disconnect(GTK_OBJECT(drawing_area), poly_button_id);

	/* Connect the motion and the button release signal handler. */
	poly_button_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "button_release_event",
	    (GtkSignalFunc) cg_poly_sh_button_release_event, NULL);

	poly_motion_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "motion_notify_event",
	    (GtkSignalFunc) cg_poly_sh_motion_notify_event, NULL);


	return (TRUE);
}

static void
cg_poly_sh_input_segment(void)
{
	/* Input the line segment. */
	poly_button_id = g_signal_connect(GTK_OBJECT(drawing_area),
	    "button_press_event",
	    (GtkSignalFunc) cg_poly_sh_button_press_event, NULL);
}

void
cg_polygon_sh(GtkWidget *w, gpointer data)
{
	g_message ("Hello, cg_polygon_sh!\n");
	poly_n = 1;
	how_many = 1;
	next_cb = cg_poly_sh_input_segment;
	cg_polygon_input_nedges();
}

void
cg_polygon_wa(GtkWidget *w, gpointer data)
{
	g_message ("Hello, cg_polygon_wa!\n");
	poly_n = 0;
	how_many = 2;
	next_cb = cg_polygon_clip_wa;
	cg_polygon_input_nedges();
}
